import json
from enum import Enum

import numpy as np
import pytest
import requests

# The conftest.py file serves as a means of providing fixtures for an entire directory.
# Fixtures defined in a conftest.py can be used by any test in that package
# without needing to import them (pytest will automatically discover them).


unit_test_data_prefix = "https://raw.githubusercontent.com/FedericoTartarini/validation-data-comfort-models/main/"


class Urls(Enum):
    ADAPTIVE_EN = "ts_adaptive_en.json"
    ADAPTIVE_ASHRAE = "ts_adaptive_ashrae.json"
    A_PMV = "ts_a_pmv.json"
    TWO_NODES = "ts_two_nodes_gagge.json"
    SOLAR_GAIN = "ts_solar_gain.json"
    ANKLE_DRAFT = "ts_ankle_draft.json"
    PHS = "ts_phs.json"
    E_PMV = "ts_e_pmv.json"
    AT = "ts_at.json"
    ATHB = "ts_athb.json"
    CLO_TOUT = "ts_clo_tout.json"
    COOLING_EFFECT = "ts_cooling_effect.json"
    VERTICAL_TMP_GRAD_PPD = "ts_vertical_tmp_grad_ppd.json"
    WBGT = "ts_wbgt.json"
    HEAT_INDEX = "ts_heat_index.json"
    NET = "ts_net.json"
    PMV_PPD = "ts_pmv_ppd.json"
    PMV = "ts_pmv.json"
    SET = "ts_set.json"
    HUMIDEX = "ts_humidex.json"
    USE_FANS_HEATWAVES = "ts_use_fans_heatwaves.json"
    UTCI = "ts_utci.json"
    WIND_CHILL = "ts_wind_chill.json"
    PET_STEADY = "ts_pet_steady.json"
    DISCOMFORT_INDEX = "ts_discomfort_index.json"


@pytest.fixture
def get_test_url():
    def _get_test_url(model_name):
        try:
            return unit_test_data_prefix + Urls[model_name.upper()].value
        except KeyError:
            return ""

    return _get_test_url


@pytest.fixture
def retrieve_data():
    def _retrieve_data(url):
        try:
            response = requests.get(url)
            if response.status_code == 200:
                return json.loads(response.text)
            else:
                response.raise_for_status()
        except requests.RequestException as e:
            print(f"Error fetching data from {url}: {e}")
        return None

    return _retrieve_data


def is_equal(a, b, tolerance=1e-6):
    if isinstance(a, np.ndarray):
        if not isinstance(b, np.ndarray):
            b = np.array(b, dtype=a.dtype)
        if a.dtype.kind in "UOS":  # U = unicode, O = objects, S = string
            return np.array_equal(a, b)
        else:
            b = np.where(b is None, np.nan, b)  # Replace None with np.nan
            # Return True if arrays are close enough, including handling of NaN values
            return np.allclose(a, b, atol=tolerance, equal_nan=True)
    elif (a is None and np.isnan(b)) or (b is None and np.isnan(a)):
        return True
    elif isinstance(a, (int, float)) and isinstance(b, (int, float)):
        # Compare scalar values with tolerance
        return np.isclose(a, b, atol=tolerance)
    else:
        return a == b


def retrieve_reference_table(get_test_url, retrieve_data, url_name):
    reference_table = retrieve_data(get_test_url(url_name))
    if reference_table is None:
        pytest.fail(f"Failed to retrieve reference table for {url_name.lower()}")
    return reference_table


def validate_result(result, expected_output, tolerance: dict):
    """Parameters
    ----------
    result this is the result of the function that is being tested
    expected_output this is the expected output of the function that is being tested
    tolerance this is the tolerance that is used to compare the result with the
    expected output

    Returns
    -------
    None

    """
    for key in expected_output:
        _expected_output = expected_output[key]

        # some functions return a dictionary or class with multiple values while
        # others return a single value
        # todo remove this once all the functions return a dictionary or class
        try:
            _result = result[key]
        except (IndexError, KeyError, TypeError):
            _result = result

        # if the key is not in the tolerance dictionary we set the tolerance to 1e-6
        if key in tolerance:
            _tolerance = tolerance[key]
        else:
            _tolerance = 1e-6

        try:
            assert is_equal(_result, _expected_output, _tolerance)
        except Exception as e:
            print(f"Expected {_expected_output}, got {_result}\nError: {str(e)}")
            raise



import pytest

from pythermalcomfort.models import humidex
from tests.conftest import Urls, retrieve_reference_table, validate_result


def test_humidex(get_test_url, retrieve_data):
    reference_table = retrieve_reference_table(
        get_test_url, retrieve_data, Urls.HUMIDEX.name
    )
    tolerance = reference_table["tolerance"]

    for entry in reference_table["data"]:
        inputs = entry["inputs"]
        outputs = entry["outputs"]
        result = humidex(**inputs)

        validate_result(result, outputs, tolerance)

    with pytest.raises(TypeError):
        humidex("25", 50)

    with pytest.raises(TypeError):
        humidex(25, "50")

    with pytest.raises(ValueError):
        humidex(tdb=25, rh=110)

    with pytest.raises(ValueError):
        humidex(25, -10)


def test_humidex_masterson():
    # todo move this to shared test
    # I got these values from
    # https://publications.gc.ca/collections/collection_2018/eccc/En57-23-1-79-eng.pdf
    result = humidex(tdb=21, rh=100, model="masterson")
    assert result.humidex == 29.3
    assert result.discomfort == "Little or no discomfort"

    result = humidex(tdb=34, rh=100, model="masterson")
    assert result.humidex == 58.4

    result = humidex(tdb=43, rh=20, model="masterson")
    assert result.humidex == 47.1

    result = humidex(tdb=30, rh=30, model="masterson")
    assert result.humidex == 31.5

    result = humidex(tdb=31, rh=55, model="masterson")
    assert result.humidex == 39.3



from pythermalcomfort.models import pmv_athb
from tests.conftest import Urls, retrieve_reference_table, validate_result


def test_athb(get_test_url, retrieve_data):
    reference_table = retrieve_reference_table(
        get_test_url, retrieve_data, Urls.ATHB.name
    )
    tolerance = reference_table["tolerance"]

    for entry in reference_table["data"]:
        inputs = entry["inputs"]
        outputs = entry["outputs"]

        result = pmv_athb(**inputs)

        validate_result(result, outputs, tolerance)



import pytest

from pythermalcomfort.models import phs
from pythermalcomfort.utilities import met_to_w_m2
from tests.conftest import Urls, retrieve_reference_table, validate_result


def test_phs(get_test_url, retrieve_data):
    reference_table = retrieve_reference_table(
        get_test_url, retrieve_data, Urls.PHS.name
    )
    tolerance = reference_table["tolerance"]

    for entry in reference_table["data"]:
        inputs = entry["inputs"]
        inputs["model"] = "7933-2004"
        outputs = entry["outputs"]
        result = phs(**inputs)

        validate_result(result, outputs, tolerance)


weight = 75
height = 1.8
a_dubois = 0.202 * (weight**0.425) * (height**0.725)


@pytest.mark.parametrize(
    "inputs, expected",
    [
        (
            dict(
                tdb=40,
                tr=40,
                rh=35,
                v=0.3,
                met=300 / met_to_w_m2 / a_dubois,
                clo=0.5,
                posture="standing",
                wme=0,
                model="7933-2023",
                duration=480,
                limit_inputs=False,
                acclimatized=100,
                round_output=False,
            ),
            dict(
                water_loss=6538,
                t_cr=37.6,
                d_lim_loss_95=280,
            ),
        ),
        (
            dict(
                tdb=35,
                tr=35,
                rh=60,
                v=0.1,
                met=300 / met_to_w_m2 / a_dubois,
                clo=0.5,
                posture="standing",
                wme=0,
                model="7933-2023",
                duration=480,
                limit_inputs=False,
                acclimatized=0,
                round_output=False,
            ),
            dict(
                water_loss=6345,
                t_cr=40.8,
                d_lim_loss_95=250,
                d_lim_t_re=62,
            ),
        ),
        pytest.param(
            dict(
                tdb=30,
                tr=54.2,
                rh=35,
                v=0.1,
                met=300 / met_to_w_m2 / a_dubois,
                clo=0.8,
                posture="standing",
                wme=0,
                a_p=0.3,
                f_r=0.85,
                model="7933-2023",
                duration=480,
                limit_inputs=False,
                acclimatized=0,
                round_output=False,
            ),
            dict(
                water_loss=6419,
                t_cr=38.7,
                d_lim_loss_95=280,
                d_lim_t_re=149,
            ),
            id="high-radiant-temp",
            marks=pytest.mark.xfail(
                reason="Known discrepancy t_cr and d_lim_t_re",
                strict=True,  # test suite fails if this case ever passes unexpectedly
            ),
        ),
        (
            dict(
                tdb=30,
                tr=30,
                rh=45,
                v=1.0,
                met=450 / met_to_w_m2 / a_dubois,
                clo=0.5,
                posture="standing",
                wme=0,
                model="7933-2023",
                duration=480,
                limit_inputs=False,
                acclimatized=0,
                round_output=False,
            ),
            dict(
                water_loss=4593,
                t_cr=38.0,
                d_lim_loss_95=400,
            ),
        ),
        (
            dict(
                tdb=35,
                tr=74.6,
                rh=30,
                v=1.0,
                met=250 / met_to_w_m2 / a_dubois,
                clo=1,
                posture="sitting",
                wme=0,
                a_p=0.2,
                f_r=0.85,
                model="7933-2023",
                duration=480,
                limit_inputs=False,
                acclimatized=100,
                round_output=False,
            ),
            dict(
                water_loss=5813,
                t_cr=37.5,
                d_lim_loss_95=310,
            ),
        ),
        pytest.param(
            dict(
                tdb=[35, 35],
                tr=74.6,
                rh=30,
                v=1.0,
                met=250 / met_to_w_m2 / a_dubois,
                clo=1,
                posture="sitting",
                wme=0,
                a_p=0.2,
                f_r=0.85,
                model="7933-2023",
                duration=480,
                limit_inputs=False,
                acclimatized=100,
                round_output=False,
            ),
            dict(
                water_loss=[5813, 5813],
                t_cr=[37.5, 37.5],
                d_lim_loss_95=[310, 310],
            ),
            id="array-input",
        ),
    ],
)
def test_2023_standard(inputs, expected):
    result = phs(**inputs)

    # check water_loss and core temperature
    assert result.water_loss == pytest.approx(expected["water_loss"], rel=0.025)
    assert result.t_cr == pytest.approx(expected["t_cr"], abs=0.3)

    assert result.d_lim_loss_95 == pytest.approx(expected["d_lim_loss_95"], abs=10)

    if "d_lim_t_re" in expected:
        assert result.d_lim_t_re == pytest.approx(expected["d_lim_t_re"], abs=10)


def test_value_acclimatized():
    with pytest.raises(ValueError):
        phs(
            tdb=40,
            tr=40,
            rh=33.85,
            v=0.3,
            met=2.58,
            clo=0.5,
            posture="standing",
            acclimatized=101,
        )

    with pytest.raises(ValueError):
        phs(
            tdb=40,
            tr=40,
            rh=33.85,
            v=0.3,
            met=2.58,
            clo=0.5,
            posture="standing",
            acclimatized=-1,
        )


def test_value_weight():
    with pytest.raises(ValueError):
        phs(
            tdb=40,
            tr=40,
            rh=33.85,
            v=0.3,
            met=2.58,
            clo=0.5,
            posture="standing",
            weight=1001,
        )

    with pytest.raises(ValueError):
        phs(
            tdb=40,
            tr=40,
            rh=33.85,
            v=0.3,
            met=2.58,
            clo=0.5,
            posture="standing",
            weight=0,
        )


def test_value_drink():
    with pytest.raises(ValueError):
        phs(
            tdb=40,
            tr=40,
            rh=33.85,
            v=0.3,
            met=2.58,
            clo=0.5,
            posture="standing",
            drink=0.5,
        )
    with pytest.raises(ValueError):
        phs(
            tdb=40,
            tr=40,
            rh=33.85,
            v=0.3,
            met=2.58,
            clo=0.5,
            posture="standing",
            drink=2,
        )



import numpy as np
import pytest

from pythermalcomfort.models import adaptive_ashrae
from pythermalcomfort.utilities import Units
from tests.conftest import Urls, retrieve_reference_table, validate_result


def test_adaptive_ashrae(get_test_url, retrieve_data):
    reference_table = retrieve_reference_table(
        get_test_url, retrieve_data, Urls.ADAPTIVE_ASHRAE.name
    )
    tolerance = reference_table["tolerance"]

    for entry in reference_table["data"]:
        inputs = entry["inputs"]
        outputs = entry["outputs"]
        units = inputs.get("units", Units.SI.value)
        result = adaptive_ashrae(
            inputs["tdb"], inputs["tr"], inputs["t_running_mean"], inputs["v"], units
        )

        validate_result(result, outputs, tolerance)


def test_ashrae_inputs_invalid_units():
    with pytest.raises(ValueError):
        adaptive_ashrae(tdb=25, tr=25, t_running_mean=20, v=0.1, units="INVALID")


def test_ashrae_inputs_invalid_tdb_type():
    with pytest.raises(TypeError):
        adaptive_ashrae(tdb="invalid", tr=25, t_running_mean=20, v=0.1)


def test_ashrae_inputs_invalid_tr_type():
    with pytest.raises(TypeError):
        adaptive_ashrae(tdb=25, tr="invalid", t_running_mean=20, v=0.1)


def test_ashrae_inputs_invalid_t_running_mean_type():
    with pytest.raises(TypeError):
        adaptive_ashrae(tdb=25, tr=25, t_running_mean="invalid", v=0.1)


def test_ashrae_inputs_invalid_v_type():
    with pytest.raises(TypeError):
        adaptive_ashrae(tdb=25, tr=25, t_running_mean=20, v="invalid")

    # Return nan values when limit_inputs=True and inputs are invalid


def test_nan_values_for_invalid_inputs():
    # Test with invalid inputs where limit_inputs=True
    result = adaptive_ashrae(
        tdb=5.0, tr=5.0, t_running_mean=5.0, v=3.0, limit_inputs=True
    )

    # Check that the comfort temperature is nan
    assert np.isnan(result.tmp_cmf)

    # Check that the acceptability flags are False
    assert result.acceptability_80 == False
    assert result.acceptability_90 == False
